
from chores import *
import time

def test_tally_simple():
    todos = Chores()
    todos.mark('one', 'good')
    todos.mark('two', 'good')
    todos.mark('three', 'bad')

    tally = todos.tally()
    assert tally.good == 2
    assert tally.good == todos.count('good')
    assert todos.marked('good') == ['one', 'two']

def test_bystatus():
    todos = Chores()
    todos.mark('one', 'good')
    todos.mark('two', 'good')
    todos.mark('three', 'bad')
    todos.mark('four', 'error')
    todos.mark('five', 'good')

    listed = todos.bystatus()
    assert listed.good == ['one', 'two', 'five']
    assert listed.bad == ['three']
    assert listed.error == ['four']

def test_access():

    todos = Chores()
    todos.mark('one', 'good')
    todos.mark('two', 'good')
    todos.mark('three', 'bad')

    assert todos['two'].status == 'good'
    todos['one'].status = 'stuff'
    assert todos['one'].status == 'stuff'
    assert todos.count('good') == 1
    assert todos.count('stuff') == 1

    # print s.report("POSTACCESS", total='TOTAL')

def test_data():
    todos = Chores()
    todos.mark('one', 'good')
    todos.mark('two', 'good')
    todos.mark('three', 'bad')

    for chore_id in todos:
        todos.data(chore_id).upper = chore_id.upper()
        todos.data(chore_id).lower = chore_id.lower()

    for chore_id in todos:
        assert todos.data(chore_id).upper == chore_id.upper()
        assert todos.data(chore_id).lower == chore_id.lower()

def test_direct_data():
    todos = Chores()
    todos.mark('one', 'good')
    todos.mark('two', 'good')
    todos.mark('three', 'bad')

    for t in todos.values():
        t.upcase = t.id.upper()

    for chore_id in todos:
        assert todos[chore_id].upcase == chore_id.upper()

def test_multi_mark():
    s = Chores()
    good = ['a', 'b', 'c']
    bad  = ['d', 'e']
    ugly = ['f', 'g']

    s.mark(good, 'good')
    s.mark(bad, 'bad')
    s.mark(ugly, 'ugly')
    assert s.marked('good') == good
    assert s.marked('bad') == bad
    assert s.marked(('good', 'bad')) == good + bad
    assert s.marked('ugly') == ugly
    assert s.marked(('good', 'ugly')) == good + ugly
    assert s.marked(('bad', 'ugly')) == bad + ugly

    assert s.marked('good', exclude='bad') == good
    assert s.marked(('good', 'bad'), exclude='bad') == good
    assert s.marked('good', exclude='ugly') == good
    assert s.marked(('good', 'bad', 'ugly'), exclude='ugly') == good + bad
    assert s.marked(('good', 'bad', 'ugly'), exclude='bad') == good + ugly


    s.mark('d', 'good')
    assert s['d'].status == 'good'
    s.mark(['a', 'c', 'f'], 'error')
    assert s.count('error') == 3
    assert s.count('good') == 2

def test_string_selection():
    s = Chores()
    good = ['a', 'b', 'c']
    bad  = ['d', 'e']
    ugly = ['f', 'g']
    s.mark(good, 'good')
    s.mark(bad, 'bad')
    s.mark(ugly, 'ugly')


    assert s.marked('good|bad') == good + bad
    assert s.marked('good|ugly') == good + ugly
    assert s.marked('bad|ugly') == bad + ugly

    assert s.marked('^bad') == good + ugly
    assert s.marked('^ugly') == good + bad
    assert s.marked('^ugly|bad') == good
    assert s.marked('^good|bad') == ugly


def test_target_statuses():
    s = Chores()
    s.mark('a', 'good')
    s.mark('b', 'good')
    s.mark('c', 'bad')
    s.mark('d', 'error')

    assert s.statuses() == 'good bad error'.split()
    assert s._target_statuses() == set('good bad error'.split())

    assert s._target_statuses('good') == set(['good'])
    assert s._target_statuses('bad') == set(['bad'])
    assert s._target_statuses('error') == set(['error'])

    assert s._target_statuses('good bad'.split()) == set(['good', 'bad'])
    assert s._target_statuses('bad error'.split()) == set(['bad', 'error'])
    assert s._target_statuses('good error'.split()) == set(['good', 'error'])

    assert s._target_statuses(exclude='good')  == set('bad error'.split())
    assert s._target_statuses(exclude='bad')   == set('good error'.split())
    assert s._target_statuses(exclude='error') == set('good bad'.split())

def test_exclude():

    s = Chores()
    s.mark('a', 'good')
    s.mark('b', 'good')
    s.mark('c', 'bad')
    s.mark('d', 'error')

    assert s.marked('good')  == s.marked(exclude='bad error'.split())
    assert s.marked('bad')   == s.marked(exclude='good error'.split())
    assert s.marked('error') == s.marked(exclude='good bad'.split())
    assert s.marked(['bad', 'error']) == s.marked(exclude='good')

def test_iteration():
    s = Chores('one two three four five'.split())
    indices = []
    names  = []
    for index, item in enumerate(s):
        s.mark(item, 'injested')
        indices.append(index)
        names.append(item)
    assert indices == [0, 1, 2, 3, 4]
    assert names  == 'one two three four five'.split()

def test_doc_example_one():

    from chores import Chores

    chores = Chores('Jones able baker charlie 8348 Smith Brown Davis'.split())

    for c in chores:
        status = 'name' if c.istitle() else 'other'
        chores.mark(c, status)


    assert chores.marked('name') == ['Jones', 'Smith', 'Brown', 'Davis']
    assert chores.count('name') == 4
    assert chores.marked(exclude='name') == [ 'able', 'baker', 'charlie', '8348']
    assert chores.count(exclude='name') == 4

    assert chores.marked('^name') == chores.marked(exclude='name')
    assert chores.count('^name') == chores.count(exclude='name')


def test_name_and_repr():
    s = Chores(name='Julian')
    assert repr(s) == "Chores('Julian')"

    s = Chores()
    r = repr(s)
    assert r.startswith("Chores('0x")
    assert r.endswith("')")

def test_len():

    s = Chores()
    assert len(s) == 0

    s = Chores(range(1))
    assert len(s) == 1

    s = Chores(range(10))
    assert len(s) == 10


def test_add_during_iteration():
    s = Chores('one two three four five'.split())
    for index, item in enumerate(s, start=1):
        if index == 3:
            s.add('error', 'ERROR')
        elif index % 2 == 0 and item != 'error':
            s.mark(item, 'even')

    assert s.statuses() == ['new', 'even', 'ERROR']
    assert s.count('new') == 3
    assert s.count('even') == 2
    assert s.count('ERROR') == 1



from chores.core import *

class Stages(object):

    """
    Represent different stages of program operation. (Experimental.)
    """

    def __init__(self, stages=None):
        self.stages = orderedstuf()
        self.open = set()
        self.begin('overall')

    def begin(self, name):
        self.open.add(name)
        if name not in self.stages:
            self.stages[name] = {}
        self.stages[name]['t0'] = time.time()

    def end(self, name):
        self.open.remove(name)
        stageinfo = self.stages[name]
        elapsed = time.time() - stageinfo['t0']
        stageinfo['elapsed'] = stageinfo.get('elapsed', 0) + elapsed
        del stageinfo['t0']

    def enter(self, name):
        return StageContext(self, name)

    def durations(self):
        """
        Return information about duration of each stage.
        """
        dur = orderedstuf()
        for n in list(self.open):
            self.end(n)
        dur.overall = self.stages['overall']['elapsed']
        for name, sinfo in self.stages.items():
            dur[name] = sinfo['elapsed']
        return dur

    def report(self):
        # close everything - is this what we want?
        for n in list(self.open):
            self.end(n)
        text = ""
        overall = self.stages['overall']['elapsed']
        for name, sinfo in self.stages.items():
            elapsed = sinfo['elapsed']
            text += "{0:8} {1:8.3f} {2:>5.1f}%\n".format(name, elapsed, elapsed/overall * 100)
        return text

class StageContext(object):
    def __init__(self, caller, name):
        self.caller = caller
        self.name = name




"""
Implement an include/exclude specification that lets strings
easily specify set inclusion/exclusion.

done|partial      : include=['done','partial']
done|previous     : include=['done','previous']
^partial|previous : exclude=['partial','previous']

Because the tags/statuses are disjoint (not overlapping)
we don't need to worry about more difficult cases specifying
both inclusions and exclusions simultaneously.
"""

import sys
_PY3 = sys.version_info[0] == 3
if _PY3:
    basestring = str

def splitwords(s, sep="|"):
    """
    Split the given string into parts on the given separator.
    Return each of the parts with possible starting and ending
    space stripped away.
    """
    return [w.strip() for w in s.strip().split(sep)]

def spec2list(spec):
    """
    Given a set specification, return a corresponding
    inclusion and exclusion list.
    """
    if not isinstance(spec, basestring):
        return spec

    spec = spec.strip()
    incl, excl = [], []
    if spec.startswith('^'):
        excl = splitwords(spec[1:])
    else:
        incl = splitwords(spec)
    return incl, excl




from chores.setspec import *


def test_splitwords():
    assert splitwords("a|b") == ['a', 'b']
    assert splitwords("a|b|c") == ['a', 'b', 'c']
    assert splitwords(" a   | b ") == ['a', 'b']
    assert splitwords(" a   | b | c ") == ['a', 'b', 'c']
    assert splitwords(" a   | b |c") == ['a', 'b', 'c']

def test_set2list():

    assert spec2list("a|b") == (['a', 'b'], [])
    assert spec2list("done|partial") == (['done', 'partial'], [])
    assert spec2list("done|previous") == (['done', 'previous'], [])
    assert spec2list("^partial|previous") == ([], ['partial', 'previous'])
    assert spec2list(" ^ partial | previous ") == ([], ['partial', 'previous'])




"""
Many programs process items of some sort, and need to track their
progress. This leads to the use of many ad hoc containers, counters,
and such. It's more effort and leads to more errors than it should.
`statustrack` helps reduce effort and errors by providing a simple,
repeatable pattern.
"""

from stuf import stuf, orderedstuf
from .counterstuf import counterstuf
from .setspec import spec2list
import time, textwrap
import sys

_PY3 = sys.version_info[0] == 3
if _PY3:
    basestring = str

class Chore(stuf):
    """
    Representation of a work item. Piggybacks a simple-to-use,
    attribute-accessible kind of ``dict``.
    """

    def mark(self, status):
        self['status'] = status

def is_sequence(arg):
    """
    Is the given object a 'proper' sequence--i.e., a sequence but not a string?
    """
    return not hasattr(arg, "strip") and (hasattr(arg, "__getitem__") or
                                          hasattr(arg, "__iter__"))

def statistics(counter):
    """
    Simple statistics getter for a dict-like structure. Designed to
    collate stats for Counter, but no reason it can't be used on any dict.
    """
    stats = stuf()
    values = counter.values()
    stats.count = len(values)
    stats.sum = sum(values)
    stats.avg = stats.sum / stats.count
    stats.min = min(values)
    stats.max = max(values)
    return stats

def dict_print(d, heading=None, wrap=True, **kwargs):
    """
    Tidy printing of dictionary elements.
    """
    lines = []

    if wrap:
        initial = kwargs.setdefault('initial_indent', '')
        kwargs.setdefault('subsequent_indent', initial + '    ')

    if heading:
        lines.append(heading)
    for key, value in d.iteritems():
        item_text = "{0}: {1}".format(key, value)
        if wrap:
            lines.extend(textwrap.wrap(item_text, **kwargs))
        else:
            lines.append(item_text)
    return '\n'.join(lines)


class Chores(object):
    """
    A class that can remember the "status" or "dispostion" of each
    item in a collection of work. Any dict can do that, right? Sure. But this
    one conveniently provides a common model and
    statistics about what happened to each item.

    The status 'total' is reserved.

    NB: Assumes each item reported on is unique and hashable.
    """

    def __init__(self, data=None, name=None, status='new'):
        """
        Start a Chores collection. Optionally give a set of initial items and a
        default status (required for iteration or creating progress reports).
        """
        self._data = orderedstuf()  # core collection, indexed by item name/id
        self._added = None  # collects items added whilst looping
        if data:
            self.add(data, status=status)
        self._name = name if name is not None else hex(id(self))
        self._added = None  # null it out again - needed

    def addone(self, key, status='new', **kwargs):
        if key in self._data:
            raise KeyError('item {0!r} already contained'.format(key))
        payload = Chore(id=key, status=status)
        payload.update(kwargs)
        self._data[key] = payload

    def add(self, items, status='new', **kwargs):
        """
        Add the given items.
        """
        items = items if is_sequence(items) else [ items ]  # in case single thing added
        newkeys = []
        if hasattr(items, 'items'): # it's a mapping
            for key, value in items.items():
                kwargs['data'] = value
                self.addone(key, status=status, **kwargs)
            newkeys = items.keys()
        else:
            for key in items:
                self.addone(key, status=status, **kwargs)
            newkeys = items
        if self._added is not None:
            self._added.extend(newkeys)

    def m_of_n(self, m, n=None, template="[m/n]"):
        """
        Return progress tracking string like '[m/n]'
        """
        n = n or len(self._data)
        return template.replace('m', str(m)).replace('n', str(n))

    def mark(self, target, status):
        """
        Mark a target item (or collection of items) as having a given status.
        """
        assert status != 'total'
        if is_sequence(target):
            for key in target:
                self.mark(key, status)
        else:
            if target in self._data:
                self._data[target].status = status
            else:
                self._data[target] = Chore(id=target, status=status)

    def _target_statuses(self, status=None, exclude=None):
        """
        Return a set of desired statuses. If status is None and exclude is
        None, consider that a wildcard and return everything. If status is a
        scalar, return any items with that status. If a sequence, any items with
        any of those statuses. Exclude is the negative selector, giving a status
        or list of statues to exclude. Exclude works best when status is None,
        so excluding against the universe of all possible statuses.
        """
        excl = set()
        if status is None:
            targets = set([ item.status for item in self._data.values() ])
        elif isinstance(status, basestring):
            incl, excl = spec2list(status)
            targets = set(incl) if incl else set([ item.status for item in self._data.values() ])
            excl = set(excl)
        elif is_sequence(status):
            targets = set(status)
        else:
            raise ValueError("not sure what to do with include {0!r}".format(status))
        if exclude or excl:
            exclude = exclude or set()
            if isinstance(exclude, basestring):
                a, b = spec2list(exclude)
                exclude = set(a) | set(b) | set(excl) # dropping possible double negative
            elif is_sequence(exclude):
                exclude = set(exclude) | excl
            else:
                exclude = exclude | excl
                # raise ValueError("not sure what to do with exclude {0!r}".format(exclude))
            targets = targets - exclude
        return targets

    def marked(self, status=None, exclude=None, justkeys=True):
        """
        Return items marked with the given status. See ``_target_statuses()`` for how
        ``status`` and ``exclude`` are interpreted.
        """
        if status is None and exclude is None:
            return list(self._data.keys() if justkeys else self._data.values())
        else:
            kvpick = (lambda k,v: k) if justkeys else (lambda k,v: v)
            target = self._target_statuses(status, exclude)
            return [ kvpick(k,v) for k,v in self._data.items() if v.status in target ]

        # Could add the ability to make the status parameter be a lambda expression

    def count(self, status=None, exclude=None):
        """
        Return items marked with the given status. See ``_target_statuses()`` for how
        ``status`` and ``exclude`` are interpreted.
        """
        return len(self.marked(status, exclude))

    def tally(self):
        """
        Tallies how many of each kind of status were seen. Return a ``counterstuf``
        (attribute-accessible version of ``collections.Counter``) of how many of
        each status marking were observed.
        """
        return counterstuf([ item.status for key, item in self._data.items() ])

    def bystatus(self, justkeys=True):
        """
        Return an ordered dictionary, indexed by status. For each status
        key, the corresponding value is an ordered list of chore ids
        that have that status. Or if justkeys is False, return a list of
        full Chore obects.
        """
        result = orderedstuf()
        for s in self.statuses():
            result[s] = []
        for key, chore in self._data.items():
            payload = key if justkeys else chore
            result[chore.status].append(payload)
        return result

    def report(self, heading=None, total='processed', order=None):
        """
        Report status. Allows an optional heading. You can choose the
        word used to represent "everything" (default: "processed")
        and the ordering of values. If order is given, it's a space-separated
        list of status names.
        """
        rep = orderedstuf()
        tally = self.tally()
        rep[total] = sum(tally.values())
        ordering = order.split() if order else sorted(tally.keys())
        for name in ordering:
            rep[name] = tally[name]
        return dict_print(rep, heading=heading, initial_indent='    ')

    # would benefit from the ability to include/exclude

    def __iter__(self):
        """
        Iterate the collection of Chores, returning a key for each.
        If items are added during processing, handle them at the end
        of the loop.
        """
        keys = list(self._data.keys())
        self._added = []
        while keys:
            for key in keys:
                yield key
            # dos-y-dos, swing her 'round!
            keys = self._added
            self._added = []
        self._added = None

        raise StopIteration

    def data(self, key):
        """
        Return an attributes-exposed dict (``Chore``) that holds data for
        each item being tracked.
        """
        return self._data.setdefault(key, Chore())

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, value):
        self._data[key] = value

    def keys(self):
        """
        Return the ids of each work item / ``Chore``.
        A list in Python 2, or a view object in Python 2.
        """
        return self._data.keys()

    def values(self):
        """
        Return ``Chore`` items.
        A list in Python 2, or a view object in Python 2.
        """
        return self._data.values()

    def items(self):
        """
        Return (key, value) tuples, where key = ``Chore`` id, value = ``Chore``.
        A list in Python 2, or a view object in Python 2.
        """
        return self._data.items()

    def __len__(self):
        return len(self._data)

    def statuses(self):
        """
        Return the unique list of status values. Implementation a little
        goofy because it tries to simulate an 'ordered set', giving statuses
        in the order seen, but with uniqueness. Thus both set and list
        mechanisms.
        """
        seen = []
        seenset = set()

        for v in self._data.values():
            status = v.status
            if status not in seenset:
                seenset.add(status)
                seen.append(status)
        return seen

    def __repr__(self):
        clsname = self.__class__.__name__
        return "{0}({1!r})".format(clsname, self._name)




    def __enter__(self):
        self.caller.begin(self.name)

    def __exit__(self, *args):
        self.caller.end(self.name)



from chores.counterstuf import counterstuf

def test_counterstuf():
    """Test counterstuf class"""
    c = counterstuf()
    c.update("this and this is this but that isn't this".split())
    c.total = sum(c.values())
    assert c.total == 9
    assert c.this == 4
    assert c.but == 1
    assert c.bozo == 0
    c2 = counterstuf().update_self("big big small big medium xlarge".split())
    assert c2.medium == 1
    assert dict(c2.most_common()) == dict([('big', 3), ('small', 1), ('medium', 1), ('xlarge', 1)])



from chores import *
import time
import pytest

def almost_equals(a, b, delta=0.01):
    """
    Is a almost equal to b?
    """
    return abs(a - b) < delta

@pytest.mark.skipif(True, reason="functionality not ready")
def test_stages():
    st = Stages('a b c'.split())
    time.sleep(0.4)
    st.begin('a')
    time.sleep(1.5)
    st.begin('b')
    time.sleep(1.8)
    st.end('a')
    st.end('b')
    with st.enter('c'):
        time.sleep(2.0)
    d = st.durations()

    assert almost_equals(d.overall, 5.7)
    assert almost_equals(d.a, 3.3)
    assert almost_equals(d.b, 1.8)
    assert almost_equals(d.c, 2.0)
